#include <windows.h>
#include <stdio.h>
#include <ddk/ntapi.h>
#include "unhook.h"
#include "ltcommon.h"
#include "ntinternals.h"

#define COMPARE_BLOCK_SIZE 12

/*
 This code was implemented for Windows 2000 SP4, Windows XP SP2 and Windows Server 2003 SP1 
 with standard memory configuration, it may crash on other systems.
*/

NTSTATUS WINAPI unhook_ZwProtectVirtualMemory(HANDLE ProcessHandle,PVOID *BaseAddress,PULONG ProtectSize,ULONG NewProtect,PULONG OldProtect,int os)
{
  NTSTATUS res;
  __asm
  (
    "push %4            \n"
    "push %3            \n"
    "push %2            \n"
    "push %1            \n"
    "push %0            \n"
    "call @org          \n"
    "mov [%6],eax       \n"
    "jmp @end           \n"

    "@org:              \n"
    "mov eax,[%5]       \n"
    "test eax,eax       \n"
    "jz @call_2k        \n"
    "dec eax            \n"
    "jz @call_xp        \n"

    "@call_2k3:         \n"
    "mov eax,0x0000008F \n"
    "mov edx,0x7FFE0300 \n"
    "call [edx]         \n"
    "ret 0x14           \n"

    "@call_xp:          \n"
    "mov eax,0x00000089 \n"
    "mov edx,0x7FFE0300 \n"
    "call [edx]         \n"
    "ret 0x14           \n"

    "@call_2k:          \n"
    "mov eax,0x00000077 \n"
    "lea edx,[esp+0x04] \n"
    "int 0x2E           \n"
    "ret 0x14           \n"

    "@end:              \n"
    :
    : "m" (ProcessHandle),
      "m" (BaseAddress),
      "m" (ProtectSize),
      "m" (NewProtect),
      "m" (OldProtect),
      "m" (os),
      "m" (res)
  );
  return res;
}


BOOL WINAPI unhook_VirtualProtect(LPVOID lpAddress,SIZE_T dwSize,DWORD flNewProtect,PDWORD lpflOldProtect)
{
  int prot=FALSE;

  prot=VirtualProtect(lpAddress,dwSize,flNewProtect,lpflOldProtect);
  if (prot) return TRUE;

  OSVERSIONINFO osvi;
  osvi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
  if (!GetVersionEx(&osvi) || (osvi.dwPlatformId!=VER_PLATFORM_WIN32_NT) || (osvi.dwMajorVersion!=5)) return FALSE;

  int os=osvi.dwMinorVersion;
  if ((os<0) || (os>2)) return FALSE;

  PVOID address=lpAddress;
  ULONG size=dwSize;
  NTSTATUS status=unhook_ZwProtectVirtualMemory(GetCurrentProcess(),&address,&size,flNewProtect,lpflOldProtect,os);
  return NT_SUCCESS(status);
}



/*
 virt_to_raw converts relative virtual offset to raw data offset, 
 if code is set to TRUE it searchs in code sections only
*/

DWORD virt_to_raw(PIMAGE_NT_HEADERS hdr,DWORD rva,int code)
{
  PIMAGE_SECTION_HEADER sec=IMAGE_FIRST_SECTION(hdr);

  for (int i=0;i<hdr->FileHeader.NumberOfSections;i++,sec++)
  {
    if ((sec->VirtualAddress<=rva) && (sec->VirtualAddress+sec->SizeOfRawData>rva) && sec->SizeOfRawData)
    {
      DWORD ret=sec->PointerToRawData+rva-sec->VirtualAddress;
      if (code & !(sec->Characteristics & (IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_CNT_CODE))) ret=0;
      return ret;
    }
  }
  return 0;
}


/*
 loc_reloc_fixup fixes up relocations in image using given relocation section and image base
*/

void loc_reloc_fixup(PIMAGE_NT_HEADERS hdr,DWORD image,PIMAGE_BASE_RELOCATION relocs,DWORD delta)
{
  ULONG blocksize;
  while (blocksize=relocs->SizeOfBlock)
  {
    ULONG pagerva=relocs->VirtualAddress;
    ULONG entries=(blocksize-sizeof(IMAGE_BASE_RELOCATION))/sizeof(WORD);

    PWORD re=(PWORD)((DWORD)relocs+sizeof(IMAGE_BASE_RELOCATION));
    while (entries>0)
    { 
      int re_type=*re >> 12;
      WORD re_ofs=*re & 0x0FFF;

      DWORD offset=image+virt_to_raw(hdr,re_ofs+pagerva,FALSE);

      switch (re_type)
      {
        case IMAGE_REL_BASED_HIGH:
          *((PWORD)offset)+=HIWORD(delta);
          break;

        case IMAGE_REL_BASED_LOW:
          *((PWORD)offset)+=LOWORD(delta);
          break;

        case IMAGE_REL_BASED_HIGHLOW:
          *((PDWORD)offset)+=delta;
          break;
      }
      re++;
      entries--;
    }
    relocs=(PIMAGE_BASE_RELOCATION)((DWORD)relocs+blocksize);
  }

  return;
}

/*
 process_library compares library memory image to its on disk image
*/

void process_library(HMODULE lib,DWORD libbase,char* dllname,int verbose)
{
  char *slash=strrchr(dllname,'\\');
  if (slash) slash++;
  else slash=dllname;

  char dir[MAX_PATH];
  lstrcpynA(dir,dllname,slash-dllname);

  HANDLE hdir=CreateFile(dir,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS ,NULL);
  if (hdir==INVALID_HANDLE_VALUE)
  {
    fprintf(stderr,"Unable to open directory \"%s\" for reading.\n",dir);
    print_last_error();
    fprintf(stderr,"\n");
    return;
  }

  OBJECT_ATTRIBUTES oa;
  IO_STATUS_BLOCK iosb;
  ANSI_STRING as;
  UNICODE_STRING us;
  RtlInitAnsiString(&as,slash);
  NTSTATUS status=RtlAnsiStringToUnicodeString(&us,&as,TRUE);
  if (!NT_SUCCESS(status))
  {
    CloseHandle(hdir);
    fprintf(stderr,"Unable to convert ansi string \"%s\" to unicode string, error code 0x%lX.\n",slash,status);
    return;
  }
  
  HANDLE ondisk;
  InitializeObjectAttributes(&oa,&us,OBJ_OPENIF | OBJ_CASE_INSENSITIVE,hdir,NULL);
  status=ZwCreateFile(&ondisk,GENERIC_READ | SYNCHRONIZE | FILE_READ_ATTRIBUTES,&oa,&iosb,0,
                      FILE_ATTRIBUTE_NORMAL,FILE_SHARE_READ | FILE_SHARE_WRITE,FILE_OPEN_IF,
                      FILE_NON_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT,NULL,0);
  RtlFreeUnicodeString(&us);
  CloseHandle(hdir);
  if (!NT_SUCCESS(status))
  {
    fprintf(stderr,"Unable to open file \"%s\" for reading, error code 0x%lX.\n",dllname,status);
    return;
  }

  if (ondisk==INVALID_HANDLE_VALUE)
  {
    fprintf(stderr,"Unable to open file \"%s\" for reading.\n",dllname);
    print_last_error();
    fprintf(stderr,"\n");
    return;
  }

  DWORD fsize=GetFileSize(ondisk,NULL);
  if (!fsize==INVALID_FILE_SIZE)
  {
    fprintf(stderr,"Unable to get library file size.\n");
    print_last_error();
    fprintf(stderr,"\n");
    CloseHandle(ondisk);
    return;
  }

  PVOID disk_image=VirtualAlloc(NULL,fsize,MEM_COMMIT,PAGE_READWRITE);
  if (!disk_image)
  {
    fprintf(stderr,"Unable to allocate %ld bytes of memory.\n",fsize);
    print_last_error();
    fprintf(stderr,"\n");
    CloseHandle(ondisk);
    return;
  }

  DWORD bytes;
  if (!ReadFile(ondisk,disk_image,fsize,&bytes,NULL))
  {
    fprintf(stderr,"Unable to read library from disk.\n");
    print_last_error();
    fprintf(stderr,"\n");
    CloseHandle(ondisk);
    VirtualFree(disk_image,0,MEM_RELEASE);
    return;
  }

  PIMAGE_DOS_HEADER dos_hdr=(PIMAGE_DOS_HEADER)disk_image;
  PIMAGE_NT_HEADERS nt_hdr=(PIMAGE_NT_HEADERS)((DWORD)disk_image+dos_hdr->e_lfanew);
  PIMAGE_DATA_DIRECTORY exp=nt_hdr->OptionalHeader.DataDirectory+IMAGE_DIRECTORY_ENTRY_EXPORT;
  PIMAGE_DATA_DIRECTORY rel=nt_hdr->OptionalHeader.DataDirectory+IMAGE_DIRECTORY_ENTRY_BASERELOC;

  ULONG exp_size=exp->Size;
  if (exp_size)
  {
    ULONG exp_addr=virt_to_raw(nt_hdr,exp->VirtualAddress,FALSE);

    ULONG rel_size=rel->Size;
    ULONG rel_addr=virt_to_raw(nt_hdr,rel->VirtualAddress,FALSE)+(DWORD)disk_image;
    ULONG delta=(ULONG)libbase-nt_hdr->OptionalHeader.ImageBase;
    if (delta && rel_size)
    {
      if (verbose) printf("Relocating library from 0x%p to 0x%p.\n",(PVOID)nt_hdr->OptionalHeader.ImageBase,lib);
      loc_reloc_fixup(nt_hdr,(DWORD)disk_image,(PVOID)rel_addr,delta);
    }

    PIMAGE_EXPORT_DIRECTORY exports=(PIMAGE_EXPORT_DIRECTORY)((DWORD)disk_image+exp_addr);
    PULONG functions = (PULONG)((DWORD)disk_image+virt_to_raw(nt_hdr,exports->AddressOfFunctions,FALSE));
    PSHORT ordinals  = (PSHORT)((DWORD)disk_image+virt_to_raw(nt_hdr,exports->AddressOfNameOrdinals,FALSE));
    PULONG names     = (PULONG)((DWORD)disk_image+virt_to_raw(nt_hdr,exports->AddressOfNames,FALSE));


    for (int i=0;i<(int)exports->NumberOfNames;i++)
    {
      ULONG ord=ordinals[i];
      if ((functions[ord]<exp->VirtualAddress) || (functions[ord]>=exp->VirtualAddress+exp_size))
      {
        DWORD rva=functions[ord];
        DWORD raw=virt_to_raw(nt_hdr,rva,TRUE);
        PVOID rawaddr=(PVOID)((DWORD)disk_image+raw);
        PVOID memaddr=(PVOID)((DWORD)lib+rva);
        DWORD hookaddr=libbase+rva;

        if (raw && ((DWORD)raw<fsize-COMPARE_BLOCK_SIZE) && memcmp(memaddr,rawaddr,COMPARE_BLOCK_SIZE))
        {
          char *fname=(char*)((DWORD)disk_image+virt_to_raw(nt_hdr,names[i],FALSE));

          UCHAR dumporg[COMPARE_BLOCK_SIZE*3];
          UCHAR dumpmod[COMPARE_BLOCK_SIZE*3];

          for (int j=0;j<COMPARE_BLOCK_SIZE;j++)
          {
            UCHAR p=*((UCHAR *)rawaddr+j);
            UCHAR q=*((UCHAR *)memaddr+j);
            snprintf(&dumporg[j*3],4,"%.2x ",p);
            snprintf(&dumpmod[j*3],4,"%.2x ",q);
          }

          if (verbose) 
            printf("  %s (0x%p) hooked!\n"
                   "    Original bytes: %s\n"
                   "    Modified bytes: %s\n",fname,(PVOID)hookaddr,dumporg,dumpmod);

          MEMORY_BASIC_INFORMATION mbi;
          if (VirtualQuery((PVOID)hookaddr,&mbi,sizeof(mbi))==sizeof(mbi))
          {
            DWORD oldprotect;
            if (unhook_VirtualProtect(mbi.BaseAddress,mbi.RegionSize,PAGE_EXECUTE_READWRITE,&oldprotect))
            {
              memcpy(memaddr,rawaddr,COMPARE_BLOCK_SIZE);
              if (verbose) printf("   Unhooked!\n");
              unhook_VirtualProtect(mbi.BaseAddress,mbi.RegionSize,oldprotect,&oldprotect);
            } else fprintf(stderr,"Unable to change protection on memory range 0x%p - 0x%p.\n",mbi.BaseAddress,(PVOID)((size_t)mbi.BaseAddress+mbi.RegionSize));
          } else fprintf(stderr,"Unable to get information about memory range 0x%p - 0x%p.\n",mbi.BaseAddress,(PVOID)((size_t)mbi.BaseAddress+mbi.RegionSize));
        }
      }
    }
  } else if (verbose) printf("This library has no export section.\n");

  VirtualFree(disk_image,0,MEM_RELEASE);

  CloseHandle(ondisk);

  if (verbose) printf("\n");
  return;
}


/*
 unhook_library calls process_library with handle and module name 
*/

void unhook_library(HMODULE module,int verbose)
{
  char name[MAX_PATH];
  GetModuleFileName(module,name,MAX_PATH);
  name[MAX_PATH-1]='\0';
  if (verbose) printf("Processing %s.\n",name);
  process_library(module,(DWORD)module,name,verbose);
  return;
}



/*
 unhook_all calls unhook_library for all modules in the process
*/

void unhook_all(int verbose)
{
  PPEB peb=get_peb();
  if (peb)
  {
    PLDR_MODULE first=(PLDR_MODULE)peb->LoaderData->InLoadOrderModuleList.Flink,current=first;
    do
    {
      if (current->BaseAddress) unhook_library(current->BaseAddress,verbose);

      current=(PLDR_MODULE)current->InLoadOrderModuleList.Flink;
    } while (current!=(PLDR_MODULE)first->InLoadOrderModuleList.Blink);
  } else fprintf(stderr,"Unable to find Process Environment Block (PEB).\n");

  return;
}



