/*

 Fake Protection Revealer 3 is a leaktest program. Please visit http://www.matousec.com/ for more information.
 
 Fake Protection Revealer (FPR or lt-fpr) is implemented to reveal fake anti-leak protection. 
 For this purpose we define the fake protection as the protection which is implemented 
 only to pass leaktests instead of fixing the real causation. FPR is implemented to 
 reveal fake protection which is based on ring 3 hooks. As stated in the article called 
 "Design of ideal personal firewall" that can be found here 
 http://www.matousec.com/projects/windows-personal-firewall-analysis/design-ideal-personal-firewall.php
 ring 3 hooks can not be used to implement security limitations for malicious processes.

 The implementation of FPR is very experimental and FPR itself should be used only 
 for testing purposes. It is not very stable because it is implemented to be a generic host 
 for other leaktests, but it is stable enough for a firewall testing. However, its technique can 
 be implemented safely, without any risks  of crashing the application. This is why it is possible 
 to implement stable malware that uses this anti-anti-leak technology.

 Firewalls that are not able to handle leaktests run by FPR are cheating on leaktests! 
 This means not only that they do not protect their users properly but they try to cover 
 their impotency and generally do offer a fake sense of security to their users.
 You can recognize the fake protection revealed by FPR easily. If you have a leaktest 
 that was not able to bypass the tested firewall and you run it using FPR, then the tested firewall 
 implements fake ring 3 protection if the leaktests succeed. Succeeding or failing leaktests run 
 by FPR that are able to bypass the tested firewall without FPR means nothing at all!

 FPR is implemented to be used with other leaktests. This means you have to obtain 
 another software to be able to test your firewall against FPR. FPR loads the given leaktest
 in its memory, unhooks all ring 3 hooks and then executes the code of the given leaktest.


 There are two versions of lt-fpr, the first (small) works on XP SP2 with 
   Breakout 1 and 2, CopyCat, FireHole, Ghost, OSfwbypass, Runner, TooLeaky, Yalta, ZAbypass

 the second (big) works on XP SP2 with 
   AWFT, Coat, CPIL, CPILSuite, DNStest, DNStester, Jumper, LeakTest, pcAudit, pcAudit 2, PCFlank, Thermite, Wallbreaker

 An extra binary was made for Surfer, it is called Surfload.exe, it is modification of the small one
 and it requires the original surfer.exe in leaktest\surfer.exe, it works on XP SP2 and is always verbose.

 FPR can not be used for BITStester.

 Updates:

   02/08/2007 - Version 3 was implemented to fight the protection of user mode hooks in Online Armor Personal Firewall 2.0.1.204.

   24/01/2007 - Version 2 was implemented to reveal a bad approach of Outpost Firewall PRO 4.0 (1005.590.123)
     to leak-tests. Many anti-leak features in prior versions of Outpost Firewall PRO 4.0 were implemented 
     improperly via ring3 hooks. FPR proved this. As a result, the vendor of Outpost implemented 
     new ring3 hook to intercept FPR, instead of solving the problem. FPR 2 shows that the solution 
     implemented in Outpost Firewall PRO 4.0 (1005.590.123) have been implemented only to pass FPR test, 
     not to solve the real problem in Outpost, which is improper usage of ring3 hooks. It is possible 
     that the next reaction of the vendor of Outpost will be another interception of this new FPR 
     using another ring3 hooks. In such case, it will be possible to modify FPR such that it will 
     be able to bypass the interception again.

*/


#undef __STRICT_ANSI__     
#include <stdio.h>
#include <windows.h>
#include <winsock.h>
#include <imagehlp.h>
#include "../common/ntinternals.h"
#include "../common/ltcommon.h"
#include "../common/unhook.h"
#include "main.h"

#define IMAGE_BASE                      (HMODULE)0x00400000
#define IMAGE_BASE_FAKE                 (HMODULE)0x00000000
#define MAX_LEAKTEST_IMAGE_SIZE         0x00080000

#ifdef BIG
int array[MAX_LEAKTEST_IMAGE_SIZE] __attribute__((section("free")));
int __attribute__((dllexport)) main(int argc,char** argv);
#endif
int __attribute__((dllimport)) __stdcall _init(void);


// verbosity: TRUE = be verbose on output
int verbose=FALSE;

// global flag for initialization purpose
int firstrun=FALSE;

void about(void)
{
  printf("Fake Protection Revealer 3 - leaktest program\n");
  printf("Windows Personal Firewall Analysis project\n");
  printf("Copyright 2006 by Matousec - Transparent security\n");
  printf("http://www.matousec.com/""\n\n");
  return;
}

void usage(void)
{
  printf("Usage: lt-fpr [options] LeaktestFileName [arg1 [arg2 [...]]]\n"
         "Options:\n"
         "  -v          be verbose on output (optional)\n\n"
         "LeaktestFileName is a relative path to a third party leaktest.\n\n"
         "Example: lt-fpr -v leaktest.exe -x something\n");
  return;
}



LONG WINAPI exception_handler(PEXCEPTION_POINTERS exception)
{
  printf("Leaktest crashed.\n");
  ExitProcess(0);
}


/*
 reloc_fixup fixes up relocations in image using given relocation section and image base
*/

void reloc_fixup(PIMAGE_NT_HEADERS hdr,DWORD image,PIMAGE_BASE_RELOCATION relocs,DWORD delta)
{
  ULONG blocksize;
  while (blocksize=relocs->SizeOfBlock)
  {
    ULONG pagerva=relocs->VirtualAddress;
    ULONG entries=(blocksize-sizeof(IMAGE_BASE_RELOCATION))/sizeof(WORD);

    PWORD re=(PWORD)((DWORD)relocs+sizeof(IMAGE_BASE_RELOCATION));
    while (entries>0)
    { 
      int re_type=*re >> 12;
      WORD re_ofs=*re & 0x0FFF;
      DWORD offset=image+re_ofs+pagerva;

      switch (re_type)
      {
        case IMAGE_REL_BASED_HIGH:
          *((PWORD)offset)+=HIWORD(delta);
          break;

        case IMAGE_REL_BASED_LOW:
          *((PWORD)offset)+=LOWORD(delta);
          break;

        case IMAGE_REL_BASED_HIGHLOW:
          *((PDWORD)offset)+=delta;
          break;
      }
      re++;
      entries--;
    }
    relocs=(PIMAGE_BASE_RELOCATION)((DWORD)relocs+blocksize);
  }

  return;
}


/*
 load_module calls LoadLibrary to load the module of a given name
 it tries to load module to its default image base
 if the function succeed the return value is HMODULE from LoadLibary call
 where the module resides
 if the function fails the return value is NULL
*/

HMODULE load_module(char *mod_name)
{
  PVOID res=NULL;

  HANDLE file=CreateFile(mod_name,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
  if (file!=INVALID_HANDLE_VALUE)
  {
    DWORD pos=SetFilePointer(file,(size_t)&((PIMAGE_DOS_HEADER)NULL)->e_lfanew,NULL,FILE_BEGIN);
    if (pos!=INVALID_SET_FILE_POINTER)
    {
      LONG e_lfanew;
      DWORD bytes;
      if (ReadFile(file,&e_lfanew,sizeof(e_lfanew),&bytes,NULL))
      {
        if (verbose)
        {
          printf("Image information:\n");
          printf("  e_lfanew = 0x%lX\n",e_lfanew);
        }

        pos=SetFilePointer(file,e_lfanew,NULL,FILE_BEGIN);
        if (pos!=INVALID_SET_FILE_POINTER)
        {
          IMAGE_NT_HEADERS nthdr;
          if (ReadFile(file,&nthdr,sizeof(nthdr),&bytes,NULL))
          {
            DWORD image_size=nthdr.OptionalHeader.SizeOfImage;
            DWORD header_size=nthdr.OptionalHeader.SizeOfHeaders;
        
            PVOID image_base=(PVOID)nthdr.OptionalHeader.ImageBase;

            if (verbose)
            {
              printf("  Image base = 0x%p\n",image_base);
              printf("  Image size = 0x%lX\n",image_size);
              printf("  Size of headers = 0x%lX\n\n",header_size);
            }

            MEMORY_BASIC_INFORMATION mbi;
            // unmap default memory location if necessary
            if (!firstrun && (VirtualQuery(image_base,&mbi,sizeof(mbi))==sizeof(mbi)))
            {
              if (verbose)
              {
                printf("Memory information for 0x%p - 0x%p:\n",mbi.AllocationBase,(PVOID)((size_t)mbi.BaseAddress+mbi.RegionSize));
                printf("  AllocationBase = 0x%p\n",mbi.AllocationBase);
                printf("  Protect = 0x%lX\n",mbi.Protect);
                printf("  Type = 0x%lX\n\n",mbi.Type);
              }

              if (mbi.AllocationBase)
              {
                if (UnmapViewOfFile(mbi.AllocationBase))
                {
                  if (verbose) printf("Memory unmapped.\n\n");
                } else
                {
                  fprintf(stderr,"Unable to unmap memory at 0x%p.\n",mbi.AllocationBase);
                  print_last_error();
                }
              }

            }

            // load image
            HMODULE image=LoadLibrary(mod_name);
            if (image)
            {
              int do_reloc=image!=image_base;
              if (verbose) printf("Image \"%s\" loaded at 0x%p%s.\n",mod_name,image,do_reloc ? " (need relocation)" : "");

              PIMAGE_DOS_HEADER doshdr=(PIMAGE_DOS_HEADER)image;
              PIMAGE_NT_HEADERS nthdr=(PIMAGE_NT_HEADERS)((size_t)image+doshdr->e_lfanew);

              // fixup page protection 
              if (verbose) printf("Modifying memory protection...\n");
              PIMAGE_SECTION_HEADER section=(PIMAGE_SECTION_HEADER)((size_t)nthdr+sizeof(IMAGE_NT_HEADERS));
              for (int i=0;i<nthdr->FileHeader.NumberOfSections;i++)
              {
                PVOID start=(PVOID)((size_t)image+section->VirtualAddress);
                size_t size=section->Misc.VirtualSize;
                PVOID end=(PVOID)((size_t)start+size);

                DWORD flags=section->Characteristics;

                if (start && size)
                {
                  DWORD oldprotect;
                  VirtualProtect(start,size,PAGE_EXECUTE_READWRITE,&oldprotect);
                }
                section++;
              }


              // fixup relocations if necessary
              PIMAGE_DATA_DIRECTORY reloc=nthdr->OptionalHeader.DataDirectory+IMAGE_DIRECTORY_ENTRY_BASERELOC;
              if (do_reloc)
              {
                if (reloc->VirtualAddress)
                {
                  PVOID reladdr=(PVOID)((size_t)image+reloc->VirtualAddress);
                  if (verbose) printf("Relocation section found at 0x%p.\n",reladdr);
                  DWORD delta=(size_t)image-(size_t)image_base;

                  reloc_fixup(nthdr,(DWORD)image,reladdr,delta);
                } else fprintf(stderr,"Unable to find relocation section.\n");
              }

              if (verbose)
              {
                section=(PIMAGE_SECTION_HEADER)((size_t)nthdr+sizeof(IMAGE_NT_HEADERS));
                for (int i=0;i<nthdr->FileHeader.NumberOfSections;i++)
                {
                  PVOID start=(PVOID)((size_t)image+section->VirtualAddress);
                  size_t size=section->Misc.VirtualSize;
                  PVOID end=(PVOID)((size_t)start+size);

                  DWORD flags=section->Characteristics;
                  char name[IMAGE_SIZEOF_SHORT_NAME+1];
                  memcpy(name,section->Name,IMAGE_SIZEOF_SHORT_NAME);
                  name[IMAGE_SIZEOF_SHORT_NAME]='\0';

                  int nlen=strlen(name);
                  for (int j=nlen;j<IMAGE_SIZEOF_SHORT_NAME;j++) name[j]=' ';

                  if (start && size)
                    printf("  #%.2d: %s range = 0x%p - 0x%p, flags = 0x%lX.\n",i,name,start,end,flags);

                  section++;
                }
                printf("\n");
              }

              // fixup imports
              PIMAGE_DATA_DIRECTORY imports=nthdr->OptionalHeader.DataDirectory+IMAGE_DIRECTORY_ENTRY_IMPORT;
              if (imports->VirtualAddress)
              {
                PVOID impaddr=(PVOID)((size_t)image+imports->VirtualAddress);
                if (verbose) printf("Import section found at 0x%p.\n",impaddr);

                if (VirtualQuery(impaddr,&mbi,sizeof(mbi))==sizeof(mbi))
                {
                  DWORD oldprotect;
                  if (VirtualProtect(mbi.BaseAddress,mbi.RegionSize,PAGE_READWRITE,&oldprotect))
                  {
                    PIMAGE_IMPORT_DESCRIPTOR importdesc=impaddr;

                    while (importdesc->Name)
                    {
                      char *name=(char*)((size_t)image+importdesc->Name);
                      HMODULE implib=LoadLibrary(name);
                      if (implib)
                      {
                        if (verbose) printf("Library \"%s\" loaded at 0x%p.\n",name,implib);

                        if (importdesc->OriginalFirstThunk)
                        {
                          PIMAGE_THUNK_DATA import_list=(PIMAGE_THUNK_DATA)((size_t)image+importdesc->OriginalFirstThunk);
                          PIMAGE_THUNK_DATA thunk_list=(PIMAGE_THUNK_DATA)((size_t)image+importdesc->FirstThunk);

                          while (import_list->u1.Ordinal)
                          {
                            // import by ordinal
                            if (IMAGE_SNAP_BY_ORDINAL(import_list->u1.Ordinal))
                            {
                              int ordinal=IMAGE_ORDINAL(import_list->u1.Ordinal);
                              thunk_list->u1.Function=(DWORD)GetProcAddress(implib,(LPCSTR)ordinal);
                              if (verbose)
                              {
                                if (thunk_list->u1.Function) printf("  Ordinal import %d found at 0x%p.\n",ordinal,(PVOID)thunk_list->u1.Function);
                                else printf("  Ordinal import %d not found.\n",ordinal);
                              }
                            } else // import by name
                            {
                              PIMAGE_IMPORT_BY_NAME nameimp=(PIMAGE_IMPORT_BY_NAME)((size_t)image+import_list->u1.AddressOfData);
                              char *proc_name=nameimp->Name;
                              thunk_list->u1.Function=(DWORD)GetProcAddress(implib,proc_name);
                              if (verbose)
                              {
                                if (thunk_list->u1.Function) printf("  Name import \"%s\" found at 0x%p.\n",proc_name,(PVOID)thunk_list->u1.Function);
                                else printf("  Name import \"%s\" not found.\n",proc_name);
                              }
                            }
                            import_list++;
                            thunk_list++;
                          }
                        } else
                        {
                          PIMAGE_THUNK_DATA thunk_list=(PIMAGE_THUNK_DATA)((size_t)image+importdesc->FirstThunk);
                          while (thunk_list->u1.Ordinal)
                          {
                            if (IMAGE_SNAP_BY_ORDINAL(thunk_list->u1.Ordinal))
                            {
                              int ordinal=IMAGE_ORDINAL(thunk_list->u1.Ordinal);
                              thunk_list->u1.Function=(DWORD)GetProcAddress(implib,(LPCSTR)ordinal);
                              if (verbose)
                              {
                                if (thunk_list->u1.Function) printf("  Ordinal import %d found at 0x%p.\n",ordinal,(PVOID)thunk_list->u1.Function);
                                else printf("  Ordinal import %d not found.\n",ordinal);
                              }
                            } else // import by name
                            {
                              PIMAGE_IMPORT_BY_NAME nameimp=(PIMAGE_IMPORT_BY_NAME)((size_t)image+thunk_list->u1.AddressOfData);
                              char *proc_name=nameimp->Name;
                              thunk_list->u1.Function=(DWORD)GetProcAddress(implib,proc_name);
                              if (verbose)
                              {
                                if (thunk_list->u1.Function) printf("  Name import \"%s\" found at 0x%p.\n",proc_name,(PVOID)thunk_list->u1.Function);
                                else printf("  Name import \"%s\" not found.\n",proc_name);
                              }
                            }
                            thunk_list++;
                          }
                        }

                        if (verbose) printf("\n");

                      } else 
                      {
                        fprintf(stderr,"Unable to load library \"%s\".\n",name);
                        print_last_error();
                      }

                      importdesc++;
                    }

                    VirtualProtect(mbi.BaseAddress,mbi.RegionSize,oldprotect,&oldprotect);
                  } else 
                  {
                    fprintf(stderr,"Unable to change memory protection for range 0x%p - 0x%p.\n",(PVOID)(mbi.BaseAddress),(PVOID)(mbi.BaseAddress+mbi.RegionSize));
                    print_last_error();
                  }
                } else 
                {
                  fprintf(stderr,"Unable to obtain information about memory address 0x%p.\n",impaddr);
                  print_last_error();
                }

              }

              res=image;
            } else 
            {
              fprintf(stderr,"VirtualAlloc failed to allocate 0x%lX bytes at 0x%p.\n",image_size,image_base);
              print_last_error();
            }

          } else
          {
            fprintf(stderr,"Unable to read %d bytes from file \"%s\".\n",sizeof(nthdr),mod_name);
            print_last_error();
          }


        } else
        {
          fprintf(stderr,"Unable to set file pointer to %ld in file \"%s\".\n",e_lfanew,mod_name);
          print_last_error();
        }
      } else 
      {
        fprintf(stderr,"Unable to read %d bytes from file \"%s\".\n",sizeof(e_lfanew),mod_name);
        print_last_error();
      }
    } else 
    {
      fprintf(stderr,"Unable to set file pointer to %d in file \"%s\".\n",(size_t)&((PIMAGE_DOS_HEADER)NULL)->e_lfanew,mod_name);
      print_last_error();
    }

    CloseHandle(file);
  } else 
  {
    fprintf(stderr,"Unable to open file \"%s\".\n",mod_name);
    print_last_error();
  }  

  return res;
}



/*
 sets EIP to the given value
*/

void set_eip(PVOID neweip)
{
  __asm
  (
    "mov eax,%0                 \n"
    "jmp eax                    \n"
    :
    : "m" (neweip)
  );
  return;
}


int main(int argc,char **argv)
{
#ifdef BIG
  firstrun=GetModuleHandle(NULL)==IMAGE_BASE;
#else
  firstrun=FALSE;
#endif
  if (!firstrun) SetUnhandledExceptionFilter(exception_handler);

  about();
  if ((argc>1) && (!strcmp(argv[1],"-v")) && !firstrun) verbose=TRUE;

  int parnameidx=verbose ? 2 : 1;
  int parargidx=parnameidx+1;

#ifdef SURFLOAD
  verbose=TRUE;
#else
  if (argc<=parnameidx)
  { 
    usage();
    return 1;
  }
#endif

  char newmodule[MAX_PATH],curmodule[MAX_PATH];
  GetModuleFileName(NULL,curmodule,MAX_PATH);
  curmodule[MAX_PATH-1]='\0';

  if (firstrun)
  {
    snprintf(newmodule,MAX_PATH,"%s.lt",curmodule);
    newmodule[MAX_PATH-1]='\0';
    if (CopyFile(curmodule,newmodule,FALSE))
    {
      if (verbose) printf("File \"%s\" copied to \"%s\".\n",curmodule,newmodule);
    } else
    {
      fprintf(stderr,"Unable to copy file from \"%s\" to \"%s\".\n",curmodule,newmodule);
      print_last_error();
      return 1;
    }
  }

  char *leaktestname=argv[parnameidx];
  char leaktestarg[MAX_PATH];
  int remain=MAX_PATH;
  char *parg=leaktestarg;

  for (int i=parargidx;i<argc;i++)
  {
    *parg='\0';
    snprintf(parg,remain,"\"%s\" ",argv[i]);
    leaktestarg[MAX_PATH-1]='\0';
    int plen=strlen(parg);
    if (!plen) break;
    parg+=strlen(parg);
    if (parg[-1]!=' ') parg[-1]='"';
    remain-=plen;
  }
  if (argc<=parargidx) leaktestarg[0]='\0';
  leaktestarg[MAX_PATH-1]='\0';
  int ltalen=strlen(leaktestarg);
  if (ltalen && (leaktestarg[ltalen-1]==' ')) leaktestarg[ltalen-1]='\0';

  int res=FALSE;

  if (firstrun) leaktestname=newmodule;

#ifdef SURFLOAD
  leaktestname="leaktest\\surfer.exe";
//  leaktestarg[0]='\0';
#endif

  HMODULE leakmod=load_module(leaktestname);
  if (leakmod)
  {
    if (verbose) printf("\nLeaktest \"%s\" loaded at 0x%p.\n",leaktestname,leakmod);

    PIMAGE_DOS_HEADER doshdr=(PIMAGE_DOS_HEADER)leakmod;
    PIMAGE_NT_HEADERS nthdr=(PIMAGE_NT_HEADERS)((size_t)leakmod+doshdr->e_lfanew);

    PVOID entry_point=(PVOID)((size_t)leakmod+nthdr->OptionalHeader.AddressOfEntryPoint);

    if (verbose) printf("Leaktest entry point found at 0x%p.\n",entry_point);

    PPEB peb=get_peb();
    if (peb)
    {
      if (verbose) printf("PEB found at 0x%p\n",peb);

      if (firstrun)
      {
        PLDR_MODULE first=(PLDR_MODULE)peb->LoaderData->InLoadOrderModuleList.Flink,current=first;
        do
        {
          if (current->BaseAddress==IMAGE_BASE)
          {
            current->BaseAddress=(PVOID)IMAGE_BASE_FAKE;
            break;
          }
          current=(PLDR_MODULE)current->InLoadOrderModuleList.Flink;
        } while (current!=(PLDR_MODULE)first->InLoadOrderModuleList.Blink);
      }

      PRTL_USER_PROCESS_PARAMETERS userpp=peb->ProcessParameters;
      if (userpp)
      {
        
        if (verbose) printf("User process parameters found at 0x%p\n\n",userpp);
        char image_path[MAX_PATH],cmd_line[MAX_PATH],cur_dir[MAX_PATH];

        if (!firstrun)
        {
          GetModuleFileName(leakmod,image_path,MAX_PATH);
          image_path[MAX_PATH-1]='\0';

          snprintf(cmd_line,MAX_PATH,"\"%s\"%s%s",image_path,leaktestarg[0] ? " " : "",leaktestarg);
          cmd_line[MAX_PATH-1]='\0';

          strncpy(cur_dir,image_path,MAX_PATH);
          cur_dir[MAX_PATH-1]='\0';
          char *slash=strrchr(cur_dir,'\\');
          if (*slash) *slash='\0';
#ifndef SURFLOAD
          SetCurrentDirectory(cur_dir);
#endif
          if (verbose)
          {
            printf("New image path = \"%s\".\n",image_path);
            printf("New command line = '%s'.\n",cmd_line);
            printf("New current directory = \"%s\".\n",cur_dir);
          }
        }

        ANSI_STRING as_cmd,as_full;
        as_cmd.Buffer   = cmd_line;
        as_full.Buffer  = image_path;
  
        as_cmd.Length   = strlen(as_cmd.Buffer);
        as_full.Length  = strlen(as_full.Buffer);

        as_cmd.MaximumLength   = as_cmd.Length;
        as_full.MaximumLength  = as_full.Length;

        size_t newinfo_size=2*(sizeof(UNICODE_STRING)+1024)+MAX_PATH;
        char *newinfo=firstrun ? NULL :malloc(newinfo_size);
        if (firstrun || newinfo)
        {
          NTSTATUS con1,con2;
          PUNICODE_STRING us_cmd=NULL,us_full=NULL;
          char *ncmd_line=NULL;

          if (!firstrun)
          {
            ncmd_line=&newinfo[2*(sizeof(UNICODE_STRING)+1024)];
            memcpy(ncmd_line,cmd_line,MAX_PATH);
            ncmd_line[MAX_PATH-1]='\0';

            us_cmd   = (PVOID)&newinfo[0*sizeof(UNICODE_STRING)];
            us_full  = (PVOID)&newinfo[1*sizeof(UNICODE_STRING)];

            char *us_start=&newinfo[2*sizeof(UNICODE_STRING)];
            us_cmd->Buffer   = (PWSTR)&us_start[0*1024];
            us_full->Buffer  = (PWSTR)&us_start[1*1024];

            us_cmd->Length   = 0;
            us_full->Length  = 0;

            us_cmd->MaximumLength   = 1024;
            us_full->MaximumLength  = 1024;
  
            con1=RtlAnsiStringToUnicodeString(us_cmd,&as_cmd,FALSE);
            con2=RtlAnsiStringToUnicodeString(us_full,&as_full,FALSE);        
          }

          if (firstrun || ((con1==STATUS_SUCCESS) && (con2==STATUS_SUCCESS)))
          {
            if (!firstrun)
            {
#ifndef SURFLOAD              
              memcpy(&userpp->ImagePathName,us_full,sizeof(UNICODE_STRING));
              memcpy(&userpp->CommandLine,us_cmd,sizeof(UNICODE_STRING));
#endif

              PVOID get_command_line_a=GetProcAddress(GetModuleHandle("kernel32"),"GetCommandLineA");
              PVOID get_command_line_w=GetProcAddress(GetModuleHandle("kernel32"),"GetCommandLineW");

              if (get_command_line_a)
              {
                unsigned char *mov=get_command_line_a;
                if (*mov==0xA1)
                {
                  PVOID *cmdptr=(PVOID*)&mov[1];
                  PVOID *cmdaddr=*cmdptr;
                  *cmdaddr=ncmd_line;
                }
              }
        
              if (get_command_line_w)
              {
                unsigned char *mov=get_command_line_w;
                if (*mov==0xA1)
                {
                  PVOID *cmdptr=(PVOID*)&mov[1];
                  PVOID *cmdaddr=*cmdptr;
                  *cmdaddr=us_cmd->Buffer;
                }
              }
            }

#ifndef SURFLOAD
            peb->ImageBaseAddress=leakmod; 
#endif
            if (verbose) printf("PEB Image Base Address set to 0x%p\n",peb->ImageBaseAddress);

            if (firstrun)
            {
              set_eip(entry_point);
            } //else
            {
              if (verbose) printf("\nUnhooking modules ...\n\n");
              unhook_all(verbose);
              if (verbose) printf("\nUnhooking done.\n\n");
              DWORD leakthread_id;
              HANDLE leakthread=CreateThread(NULL,0,entry_point,NULL,0,&leakthread_id);
              if (leakthread)
              {
                res=TRUE;
                if (verbose) printf("Leaktest thread runs, ThreadId = 0x%lX.\n\n\n",leakthread_id);
                WaitForSingleObject(leakthread,INFINITE);
                // the code of original module usually calls ExitProcess and so this wait never returns 
                if (verbose) printf("Leaktest thread terminated.\n");
                CloseHandle(leakthread);
              } else 
              {
                fprintf(stderr,"Unable to create new thread.\n");
                print_last_error();
              }
            }
          } else fprintf(stderr,"Unable to convert ansi strings to unicode strings using RtlAnsiStringToUnicodeString.\n"
                                "Return values were 0x%lX, 0x%lX.\n",con1,con2);
        } else fprintf(stderr,"Unable to allocate %d bytes.\n",newinfo_size);
      } else fprintf(stderr,"Unable to find user process parameters.\n");
    } else fprintf(stderr,"Unable to find Process Environment Block (PEB).\n");
  } else fprintf(stderr,"Loading leaktest \"%s\" failed.\n",leaktestname);

  if (res) printf("\nTEST SUCCESSFUL!\n");
  else printf("\nTEST FAILED!\n");

  // static loading of important DLLs
  if (res+1==10)
  {
    ShellExecuteEx(NULL);
    MessageBox(0,NULL,NULL,0);
    PlaySound(NULL,NULL,0);
    _init();
    closesocket(0);
  }
  return (res ? 0 : 1);
}
